<?php
  /**
   * Třída sloužící k získávání informací o bezpečnostních incidentech, k přidávání nových incidentů,
   * k úpravě těch existujících a k dalším souvisejícím operacím.
   *
   * @author Martin Šebela
   */
  class IncidentModel extends FormModel {
    /**
     * @var string      Název tabulky v databázi, se kterou je instance svázána.
     */
    private $dbTableName = 'blck_incidents';

    /**
     * @var int         Číslo lístku (ticketu) se stížností.
     */
    protected $idParentTicket;

    /**
     * @var array       Jazyk, v jakém je vedena komunikace se stěžovatelem v lístku se stížností.
     */
    protected $parentTicketLang;

    /**
     * @var int         ID typu incidentu.
     */
    protected $idType;

    /**
     * @var string      IP adresa, která je předmětem incidentu.
     */
    protected $blockedIp;

    /**
     * @var string      MAC adresa, která je předmětem incidentu.
     */
    protected $blockedMac;

    /**
     * @var string      Hostname zařízení, které je předmětem incidentu.
     */
    protected $blockedHostname;

    /**
     * @var string      Datum a čas, kdy k incidentu došlo.
     */
    protected $incidentDatetime;

    /**
     * @var string      Příjemce zprávy o blokaci.
     */
    protected $blockingRecipient;

    /**
     * @var string      Předmět zprávy o blokaci.
     */
    protected $blockingSubject;

    /**
     * @var string      Obsah zprávy o blokaci.
     */
    protected $blockingMessage;

    /**
     * @var string      Příjemce zprávy (stěžovatel) o vyřešení incidentu.
     */
    protected $notifyRecipient;

    /**
     * @var string      Obsah zprávy pro stěžovatele o vyřešení incidentu.
     */
    protected $notifyMessage;

    /**
     * @var string      Obsah zprávy o odblokování.
     */
    protected $unblockingMessage;

    /**
     * @var array       Uživatel, jehož se incident týká (resp. uživatel vybraný v kroku identifikace).
     */
    protected $user;

    /**
     * @var array       Jazykové mutace, v nichž bude e-mail odeslaný uživateli/lokálnímu správci (původci incidentu).
     */
    protected $userLang;

    /**
     * @var int         Následující krok, na který má být uživatel přesměrován při řešení incidentu.
     */
    protected $nextStep = 1;

    /**
     * @var int         Aktuální krok při řešení incidentu.
     */
    protected $currentStep = 1;


    /**
     * Nastaví u incidentu krok, který je právě zobrazen ve formuláři.
     *
     * @param int $step                Aktuální krok ve formuláři
     */
    public function setCurrentStep($step) {
      $this->currentStep = $step;
    }


    /**
     * Smaže data z paměti dosud vložená uživatelem, a vyprázdní tak formulář.
     */
    public function reset($action) {
      $this->currentStep = 1;
      $this->nextStep = 1;

      unset($_SESSION['tempIncident' . $action]);
    }


    /**
     * Zpracuje vstup od uživatele na konkrétní krok při řešení bezpečnostního incidentu
     * a jako výstup vrátí následující krok, na který má být uživatel přesměrován.
     *
     * @param int $step                Číslo kroku, které je ve formuláři zpracováváno.
     * @param string $action           Typ akce (přidání nového incidentu nebo editace).
     * @param null $idIncident         ID incidentu (pokud se jedná o editaci) [nepovinné].
     * @return int                     Číslo kroku, na který má být uživatel přesměrován.
     * @throws UserError               Výjimka obsahující textovou informaci o chybě pro uživatele.
     */
    public function makeIncidentStep($step, $action, $idIncident = null) {
      $incident = $this->getTempIncident($action);

      if (empty($incident)) {
        $incident = $this->getIncident($idIncident);
      }

      // Nalezení vhodného modulu, který je schopen provádět s danou IP adresu operace (hledat v logu, blokovat, ...).
      $blocker = get_blocker_by_hostname((!empty($incident['blocked_ip'])) ? $incident['blocked_ip'] : $this->blockedIp);

      if ($blocker == null) {
        throw new UserError('Pro uvedenou IP adresu nebyl nalezen žádný modul ke správě.', MSG_ERROR);
      }

      if ($this->currentStep == 1) {
        // Založení nového incidentu.
        $incident = $this->confirmNewIncident($blocker, $incident);
        $this->nextStep = $this->currentStep + 1;
      }
      else if ($this->currentStep == 2) {
        // Identifikace problémového uživatele nebo zařízení.
        $incident = $this->confirmIdentification($incident);

        $this->nextStep = $this->currentStep + 1;
      }
      else if ($this->currentStep == 3) {
        // Blokace konkrétního zařízení nebo uživatele.
        $incident = $this->confirmBlocation($blocker, $incident);

        if (!empty($incident['id_parent_ticket'])) {
          $this->nextStep = $this->currentStep + 1;
        }
        else {
          // Pokud nebyl vyplněn rodičovský lístek, přeskočit krok "upozornění stěžovatele".
          $this->nextStep = $this->currentStep + 2;
        }
      }
      else if ($this->currentStep == 4) {
        $this->nextStep = $this->currentStep + 1;

        // Upozornění stěžovatele na vyřešení incidentu.
        $incident = $this->confirmNotify($blocker, $incident);
      }
      else if ($this->currentStep == 5) {
        $this->nextStep = $this->currentStep + 1;

        // Odblokování dříve zablokovaného uživatele nebo zařízení.
        $incident = $this->confirmUnblocking($blocker, $incident);
      }

      $incident['last_step'] = $this->nextStep;

      $this->dbRecordData = $incident;
      $_SESSION['tempIncident' . $action] = $this->dbRecordData;

      return $this->nextStep;
    }


    /**
     * Potvrdí první krok řešení bezpečnostního incidentu - doplnění základních informací o bezpečnostním
     * incidentu a připraví data pro formulář v následujícím kroku.
     *
     * @param IBlocker $blocker        Blockátor, který umožňuje pracovat s daným rozsahem sítě.
     * @param array $incident          Data o incidentu.
     * @return array                   Nově doplněná data o incidentu.
     * @throws UserError               Výjimka obsahující textovou informaci o chybě pro uživatele.
     */
    private function confirmNewIncident($blocker, $incident) {
      $this->blockedMac = mb_strtolower($this->blockedMac);
      $this->blockedHostname = gethostbyaddr($this->blockedIp);
      $this->incidentDatetime = get_formatted_datetime($this->incidentDatetime);

      $incident['blocked_ip'] = $this->blockedIp;
      $incident['blocked_mac'] = $this->blockedMac;
      $incident['blocked_hostname'] = $this->blockedHostname;
      $incident['incident_datetime'] = $this->incidentDatetime;
      $incident['id_type'] = $this->idType;
      $incident['id_parent_ticket'] = (!empty($this->idParentTicket)) ? $this->idParentTicket : null;

      // Příprava dat pro krok 2 (identifikace uživatele/zařízení).
      $evidence = $blocker->find($this->blockedIp, $this->blockedMac, $this->incidentDatetime);

      // Pokud se podařilo najít záznamy v některém z logů...
      if (!empty($evidence['logs'])) {
        // Připravit barevně zvýrazněné logy.
        $incident['incident_evidence'] = get_array_items_as_newline_string($evidence['logs']);
        $incident['incident_evidence_formatted'] = $this->markIdentifiersInLogs(
          $this->blockedIp, $evidence['macAddresses'], $evidence['hostnames'], $evidence['users'], $evidence['logs']
        );

        $incident['user_detail'] = [];

        $ldap = new LdapModel();

        // Zjištění dalších informací o uživateli z LDAP (pokud je uživatel z dané organizace).
        foreach ($evidence['users'] as &$user) {
          if (is_user_from_organization($user)) {
            $user = get_email_part($user, 'username');

            $incident['user_detail'][] = $ldap->getUserCNByUsername($user);
            $email = $ldap->getEmailByUsername($user);

            // E-mail se nemusí podařit z LDAP dostat (např. u hostovských kont).
            if (!empty($email)) {
              $user = $email;
            }
          }
          else {
            $incident['user_detail'][] = '';
          }
        }

        // Pokud se nepodařilo nalézt žádného uživatele nebo lokálního správce, pak použít
        // výchozí e-mail bezpečnostního týmu, aby bylo možné provést blokaci.
        if (empty($evidence['users'])) $evidence['users'][] = CSIRT_ABUSE_CONTACT;

        // Doplnění identifikovaného uživatele, MAC adresy a hostname zařízení
        // a typu sítě na základě konkrétního Blockátoru.
        $incident['identifiedUsers'] = $evidence['users'];
        $incident['macAddresses'] = $evidence['macAddresses'];
        $incident['hostnames'] = $evidence['hostnames'];
        $incident['network'] = $blocker->getNetworkName();

        // Výchozí nastavení jazykových mutací u šablon pro e-maily.
        if (empty($incident['user_lang'])) $incident['user_lang'] = [INCIDENT_USER_DEFAULT_LANG];
        if (empty($incident['parent_ticket_lang'])) $incident['parent_ticket_lang'] = [INCIDENT_COMPLAINANT_DEFAULT_LANG];

        // Datum a čas, kdy došlo k identifikaci.
        $incident['date_identification'] = date('Y-m-d H:i:s');
      }
      else {
        $this->nextStep--;
        throw new UserError('V žádném z logů se nepodařilo nalézt uvedenou IP adresu v daném čase.', MSG_ERROR);
      }

      return $incident;
    }


    /**
     * Potvrdí druhý krok řešení bezpečnostního incidentu - identifikaci problémového uživatele/zařízení
     * a připraví data pro formulář v následujícím kroku.
     *
     * @param array $incident          Data o incidentu.
     * @return array                   Nově doplněná data o incidentu.
     */
    private function confirmIdentification($incident) {
      // Teď už nastavíme vybraného uživatele (v jednom incidentu lze blokovat pouze jednoho uživatele/zařízení).
      $incident['blocked_user'] = $this->user[0];
      $incident['blocking_recipient'] = implode(',', $this->user);

      $this->user = $this->user[0];

      // Pokud uživatel nevyplnil MAC adresu, ale nyní ji již známe z logů, tak ji zpětně doplníme.
      if (empty($incident['blocked_mac']) && isset($incident['macAddresses'][0])) {
        $incident['blocked_mac'] = $incident['macAddresses'][0];
      }

      // Zjištění zvolených jazykových mutací.
      $incident['user_lang'] = $this->userLang;
      $incident['parent_ticket_lang'] = $this->parentTicketLang;

      // Načtení zprávy o blokaci na základě zjištěné jazykové mutace.
      $message = $this->getMessages($incident['id_type'], 3, $this->user, $incident['network'], $this->userLang, false);
      $incident['blocking_subject'] = $message['subject'];
      $incident['blocking_message'] = $message['body'];

      // Načtení zprávy o odblokování na základě zjištěné jazykové mutace.
      $incident['unblocking_message'] = $this->getMessages($incident['id_type'], 5, $this->user, $incident['network'], $this->userLang);

      return $incident;
    }


    /**
     * Potvrdí třetí krok řešení bezpečnostního incidentu - blokaci konkrétního uživatele, nebo zařízení
     * a připraví data pro formulář v následujícím kroku.
     *
     * @param IBlocker $blocker        Blockátor, který umožňuje pracovat s daným rozsahem sítě.
     * @param array $incident          Data o incidentu.
     * @return array                   Nově doplněná data o incidentu.
     * @throws UserError               Výjimka obsahující textovou informaci o chybě pro uživatele.
     */
    private function confirmBlocation($blocker, $incident) {
      $incident['blocking_recipient'] = $this->blockingRecipient;
      $incident['blocking_subject'] = $this->blockingSubject;
      $incident['blocking_message'] = $this->blockingMessage;

      if (!empty($incident['id_parent_ticket'])) {
        // Zjištění kontaktu (e-mailu) na žadatele (stěžovatele) z rodičovského lístku z RT systému.
        $parentTicket = TicketingSystem::getTicketMetadata($incident['id_parent_ticket']);
        $incident['notify_recipient'] = (isset($parentTicket['Requestors'])) ? $parentTicket['Requestors'] : '';

        // Zjištění obsahu zprávy pro stěžovatele.
        $incident['notify_message'] = $this->getMessages($incident['id_type'], 4, $incident['blocked_user'], $incident['network'], $incident['parent_ticket_lang']);

        $nextStep = $this->currentStep + 1;
      }
      else {
        // Pokud nebyl vyplněn rodičovský lístek na stěžovatele, pak přeskočit daný krok a přejít rovnou k odblokaci.
        $nextStep = $this->currentStep + 2;
      }

      // Přidání lístku do RT a zjištění jeho ID.
      $incident['id_child_ticket'] = TicketingSystem::addNewTicket($incident['blocking_recipient'], $incident['blocking_subject'], $incident['blocking_message']);

      // Přidání bezpečnostního incidentu do Blockatoru.
      $incident['id_incident'] = $this->insertIncident($incident, $nextStep);
      $this->insertLogAction($incident['id_incident'], $this->currentStep - 1, $incident['date_identification']);

      // Pokud byl vyplněn lístek stěžovatele, přidat v RT systému vazbu na tento lístek.
      if (!empty($incident['id_parent_ticket'])) {
        TicketingSystem::addDependsToTicket($incident['id_child_ticket'], $incident['id_parent_ticket']);
      }

      // Vytvoření blokované identity, která je předávána konkrétnímu modulu (Blockátoru).
      $entity = new BlockedIdentity(
        $incident['blocked_ip'], $incident['blocked_mac'], $incident['blocked_user'],
        $incident['id_child_ticket'], $incident['id_incident'],
        IncidentTypeModel::getNameIncidentType($incident['id_type'])
      );

      if (!$blocker->block($entity)) {
        throw new UserError('Uživatele nebo problémové zařízení se nepodařilo zablokovat z důvodu nefunkčnosti externích systémů.', MSG_ERROR);
      }

      // Komentář do RT o úspěšně provedené blokaci.
      TicketingSystem::commentTicket($incident['id_child_ticket'], $blocker->getRTBlockedMessage($entity));

      return $incident;
    }


    /**
     * Potvrdí čtvrtý, ovšem nepovinný krok řešení bezpečnostního incidentu - notifikaci stěžovatele
     * a připraví data pro formulář v následujícím kroku.
     *
     * @param IBlocker $blocker        Blockátor, který umožňuje pracovat s daným rozsahem sítě.
     * @param array $incident          Data o incidentu.
     * @return array                   Nově doplněná data o incidentu.
     * @throws UserError               Výjimka obsahující textovou informaci o chybě pro uživatele.
     */
    private function confirmNotify($blocker, $incident) {
      $incident['notify_recipient'] = $this->notifyRecipient;
      $incident['notify_message'] = $this->notifyMessage;

      if (!empty($this->notifyMessage)) {
        // Odpověď do lístku stěžovatele (pokud byl vyplněn obsah zprávy, není totiž povinný).
        TicketingSystem::replyTicket($incident['id_parent_ticket'], $this->notifyMessage);
      }
      else {
        // V opačném případě se stěžovateli pošle pouze automatická zpráva o vyřešení incidentu.
        TicketingSystem::changeTicketStatus($incident['id_parent_ticket'], 'resolved');
      }

      // Data, která se aktualizují v databázi.
      $edited = [
        'notify_recipient' => $this->notifyRecipient,
        'notify_message' => $this->notifyMessage
      ];

      $this->updateIncident($incident['id_incident'], $edited);

      return $incident;
    }


    /**
     * Potvrdí poslední, pátý krok řešení bezpečnostního incidentu - odblokování uživatele nebo zařízení.
     *
     * @param IBlocker $blocker        Blockátor, který umožňuje pracovat s daným rozsahem sítě.
     * @param array $incident          Data o incidentu.
     * @return array                   Nově doplněná data o incidentu.
     * @throws UserError               Výjimka obsahující textovou informaci o chybě pro uživatele.
     */
    private function confirmUnblocking($blocker, $incident) {
      $incident['unblocking_message'] = $this->unblockingMessage;

      // Odpověď problémovému uživateli (popř. lokálnímu správci) o odblokování a vyřešení incidentu.
      TicketingSystem::replyTicket($incident['id_child_ticket'], $this->unblockingMessage);
      TicketingSystem::changeTicketStatus($incident['id_child_ticket'], 'resolved');

      $entity = new BlockedIdentity(
        $incident['blocked_ip'], $incident['blocked_mac'], $incident['blocked_user'],
        $incident['id_child_ticket'], $incident['id_incident'],
        IncidentTypeModel::getNameIncidentType($incident['id_type'])
      );

      if (!$blocker->unblock($entity)) {
        throw new UserError('Uživatele nebo problémové zařízení se nepodařilo odblokovat z důvodu nefunkčnosti externích systémů.', MSG_ERROR);
      }

      // Data, která se aktualizují v databázi.
      $edited = [
        'unblocking_message' => $this->unblockingMessage
      ];

      $this->updateIncident($incident['id_incident'], $edited);

      return $incident;
    }


    /**
     * Vrátí připravenou, personalizovanou zprávu pro uživatele (popř. stěžovatele), a to na základě předaných parametrů.
     *
     * @param int $idType              ID typu bezpečnostního incidentu.
     * @param int $idStatus            ID stavu, pro nějž je zpráva vyžadována.
     * @param string $username         Uživatelské jméno, vůči němuž se má zpráva personalizovat.
     * @param string $network          Název sítě, jíž se incident týká.
     * @param array $languages         Jazykové mutace, v nichž má být zpráva vrácena.
     * @param bool $withoutSubject     Určuje, zdali se má vrátit pouze tělo zprávy bez předmětu (výchozí je TRUE).
     * @return array|string            Personalizovaná zpráva (buď formou pole s klíči nebo jako řetězec).
     */
    private function getMessages($idType, $idStatus, $username, $network, $languages, $withoutSubject = true) {
      $message['subject'] = '';
      $message['body'] = '';

      foreach ($languages as $i => $language) {
        $databaseMessage = EmailTemplatesModel::getEmailTemplateByStatus($idType, $idStatus, $language);

        if (!$withoutSubject) {
          $message['subject'] .= EmailTemplatesModel::getPersonalizedContent($databaseMessage['subject'], $network, $username);
        }

        $message['body'] .= EmailTemplatesModel::getPersonalizedContent($databaseMessage['body'], $network, $username);

        if (count($languages) > 1 && $i != count($languages) - 1) {
          if (!$withoutSubject) {
            $message['subject'] .= INCIDENT_MESSAGE_SUBJECT_DELIM;
          }

          $message['body'] .= INCIDENT_MESSAGE_BODY_DELIM;
        }
      }

      if ($withoutSubject) {
        $message = $message['body'];
      }

      return $message;
    }


    /**
     * Vrátí a načte aktuálně zpracovávaný bezpečnostní incident.
     *
     * @return array|null              Pole s informacemi o právě zpracovávaném bezpečnostním incidentu nebo NULL.
     */
    public function getTempIncident($action) {
      $data = null;

      if (isset($_SESSION['tempIncident' . $action])) {
        $this->dbRecordData = $_SESSION['tempIncident' . $action];
        $data = $this->dbRecordData;
      }

      return $data;
    }


    /**
     * Uloží do instance a zároveň vrátí (z databáze) informace o zvoleném bezpečnostním incidentu.
     *
     * @param int $id                  ID bezpečnostního incidentu.
     * @return array                   Pole obsahující informace o incidentu.
     */
    public function getIncident($id) {
      $incident = Database::querySingle('
              SELECT `id_incident`, `id_by_user`, `id_status`, `id_type`, `id_parent_ticket`, `id_child_ticket`, `blocked_user`, `blocked_ip`, `blocked_mac`, `blocked_hostname`, `network`, `incident_datetime`, `incident_evidence`, `user_lang`, `blocking_recipient`, `blocking_subject`, `blocking_message`, `unblocking_message`, `notify_recipient`, `notify_message`, `user_lang`, `parent_ticket_lang`, `date_added`
              FROM `blck_incidents`
              WHERE `id_incident` = ?
              AND `visible` = 1'
      , $id);
      $this->dbRecordData = $incident;

      // Načtení dat z databáze tak, aby to vyhovovalo jednotným podmínkám pro každý modul ve formuláři.
      if (!empty($incident)) {
        // Připravit zvýrazněný log na základě dat z databáze.
        $this->dbRecordData['incident_evidence_formatted'] = $this->markIdentifiersInLogs(
          $incident['blocked_ip'],
          [$incident['blocked_mac']],
          [$incident['blocked_hostname']],
          [$incident['blocked_user']],
          [$incident['incident_evidence']]
        );

        $ldap = new LdapModel();

        // Informace o identifikovaném uživateli.
        $this->dbRecordData['identifiedUsers'] = [$incident['blocked_user']];
        $this->dbRecordData['user_detail'][] = $ldap->getUserCNByUsername(get_email_part($incident['blocked_user'], 'username'));
        $this->dbRecordData['macAddresses'] = [$incident['blocked_mac']];
        $this->dbRecordData['hostnames'] = [$incident['blocked_hostname']];

        // Záznamy o použitých jazykových mutacích v odeslaných zprávách.
        $this->dbRecordData['user_lang'] = explode(',', $incident['user_lang']);
        $this->dbRecordData['parent_ticket_lang'] = explode(',', $incident['parent_ticket_lang']);
      }

      return $this->dbRecordData;
    }



    /**
     * Vrátí seznam všech bezpečnostních incidentů z databáze.
     *
     * @return mixed                   Pole bezpečnostních incidentů a informace o každém z nich.
     */
    public static function getIncidents() {
      $data = Database::queryMulti('
              SELECT `id_incident`, blck_incidents.id_by_user, `id_status`, `id_type`, `id_parent_ticket`, `id_child_ticket`, `blocked_user`, `blocked_ip`, `blocked_mac`, `blocked_hostname`, `network`,
              `username`,
              `id_incident_status`, `title`, blck_incidents_status.color AS status_color,
              `name`, blck_incidents_types.color,
              DATE_FORMAT(blck_incidents.incident_datetime, "%e. %c. %Y %H:%i") AS incident_datetime,
              DATE_FORMAT(blck_incidents.date_added, "%e. %c. %Y %H:%i") AS date_added,
              DATE_FORMAT(TIMEDIFF(blck_incidents.date_added, blck_incidents.incident_datetime), "%k:%i") AS reaction_speed
              FROM `blck_incidents`
              JOIN `blck_users`
              ON blck_incidents.id_by_user = blck_users.id_user
              JOIN `blck_incidents_status`
              ON blck_incidents.id_status = blck_incidents_status.id_incident_status
              JOIN `blck_incidents_types`
              ON blck_incidents.id_type = blck_incidents_types.id_incident_type
              WHERE blck_incidents.visible = 1
              ORDER BY `id_incident` DESC
      ');

      foreach ($data as &$incident) {
        $incident['user_from_organization'] = is_user_from_organization($incident['blocked_user']);

        // V případě, že je uživatel z dané organizace, pak nevracet část za @ (doménu).
        if ($incident['user_from_organization']) {
          $incident['blocked_user'] = get_email_part($incident['blocked_user'], 'username');
        }

        // Rychlost reakce na incident.
        if (mb_substr($incident['reaction_speed'], 0, 3) == '00:') {
          $incident['reaction_speed'] = mb_substr($incident['reaction_speed'], 3);
        }

        $incident['reaction_speed_color'] = self::getReactionSpeedColor($incident['reaction_speed']);

        $status = self::getLastIncidentStatus($incident['id_incident']);
        $incident['id_status'] = $status['id_incident_status'];
        $incident['title'] = $status['title'];
      }

      return $data;
    }


    /**
     * Vrátí celkový počet incidentů, případně pouze počet vyřešených nebo nevyřešených incidentů.
     *
     * @param string $type             Nepovinný parametr určující, zdali se má vrátit počet vyřešených ("resolved")
     *                                 nebo počet nevyřešených ("unresolved") incidentů.
     * @return int                     Počet incidentů vyhovujících podmínce.
     */
    public static function getIncidentsCount($type = null) {
      $whereFilter = '';

      if ($type != null) {
        switch ($type) {
          case 'resolved':
            $whereFilter = 'AND id_status = 6'; break;
          case 'unresolved':
            $whereFilter = 'AND id_status != 6'; break;
        }
      }

      return Database::queryCount('
              SELECT COUNT(*)
              FROM `blck_incidents`
              WHERE `visible` = 1
              ' . $whereFilter . '
      ');
    }


    /**
     * Vrátí poslední stav, ke kterému u konkrétního bezpečnostního stavu došlo.
     *
     * @param int $idIncident          ID bezpečnostního incidentu, u něhož se má stav zjišťovat.
     * @return array                   Informace o stavu.
     */
    public static function getLastIncidentStatus($idIncident) {
      return Database::querySingle('
              SELECT blck_incidents_status_log.id_incident_status, `date_added`,
              `title`
              FROM `blck_incidents_status_log`
              JOIN `blck_incidents_status`
              ON blck_incidents_status_log.id_incident_status = blck_incidents_status.id_incident_status
              WHERE `id_incident` = ?
              ORDER BY `id_incident_status` DESC'
        , $idIncident);
    }


    /**
     * Vloží do databáze nový bezpečnostní incident.
     */
    public function insertIncident($incident, $nextStep) {
      $userTicketLang = implode(',', $incident['user_lang']);
      $parentTicketLang = (!empty($incident['parent_ticket_lang'])) ? implode(',', $incident['parent_ticket_lang']) : '';

      $newIncident = [
        'id_by_user' => PermissionsModel::getUserId(),
        'id_status' => $nextStep,
        'id_type' => $incident['id_type'],

        'id_parent_ticket' => $incident['id_parent_ticket'],
        'id_child_ticket' => $incident['id_child_ticket'],

        'blocked_user' => $incident['blocked_user'],
        'blocked_ip' => $incident['blocked_ip'],
        'blocked_mac' => $incident['blocked_mac'],
        'blocked_hostname' => $incident['blocked_hostname'],
        'network' => $incident['network'],

        'incident_datetime' => $incident['incident_datetime'],
        'incident_evidence' => $incident['incident_evidence'],

        'user_lang' => $userTicketLang,
        'parent_ticket_lang' => $parentTicketLang,

        'blocking_recipient' => $incident['blocking_recipient'],
        'blocking_subject' => $incident['blocking_subject'],
        'blocking_message' => $incident['blocking_message'],
        'unblocking_message' => $incident['unblocking_message'],

        'date_added' => date('Y-m-d H:i:s')
      ];

      if (isset($incident['notify_recipient']) && isset($incident['notify_message'])) {
        $newIncident['notify_recipient'] = $incident['notify_recipient'];
        $newIncident['notify_message'] = $incident['notify_message'];
      }

      Logger::info('Vložení nového bezpečnostního incidentu.', $newIncident);

      Database::insert($this->getDbTableName(), $newIncident);
      $incidentId = Database::getLastInsertId();

      $this->insertLogAction($incidentId, $this->currentStep);

      return $incidentId;
    }


    /**
     * Upraví zvolený bezpečnostní incident.
     *
     * @param int $id                  ID bezpečnostního incidentu.
     * @param array $incident          Upravovaná data u incidentu.
     */
    public function updateIncident($id, $incident) {
      $incident['id_status'] = $this->nextStep;

      Logger::info('Úprava existujícího bezpečnostního incidentu.', $incident);

      Database::update(
        $this->getDbTableName(),
        $incident,
        'WHERE `id_incident` = ? AND `visible` = 1',
        $id
      );

      $this->insertLogAction($id, $this->currentStep);

      // Při odblokování dojde zároveň k vyřešení incidentu,
      // tzn. vložit do databáze oba záznamy.
      if ($this->nextStep == INCIDENT_FINAL_STEP_ID) {
        $this->insertLogAction($id, INCIDENT_FINAL_STEP_ID);
      }
    }


    /**
     * Vloží do databáze nový záznam o změně stavu u konkrétního incidentu.
     *
     * @param int $idIncident          ID bezpečnostního incidentu
     * @param int $idStatus            ID stavu
     * @param null $datetime           Vlastní datum a čas, kdy k akci došlo (nepovinný parametr).
     */
    private function insertLogAction($idIncident, $idStatus, $datetime = null) {
      $logRecord = [
        'id_incident' => $idIncident,
        'id_incident_status' => $idStatus,
        'date_added' => date('Y-m-d H:i:s')
      ];

      if ($datetime != null) {
        $logRecord['date_added'] = $datetime;
      }

      Database::insert('blck_incidents_status_log', $logRecord);
    }


    /**
     * Vrátí seznam všech možných stavů (resp. kroků) při řešení bezpečnostního incidentu.
     * V případě předání konkrétního ID incidentu dojde ke zjištění i data změny u každého ze stavů.
     *
     * @param int|false $idIncident    ID incidentu, pro který se má vracet seznam stavů včetně data
     *                                 změny nebo FALSE (výchozí parametr).
     * @return mixed                   Pole všech stavů (případně včetně data změny).
     */
    public static function getIncidentStatuses($idIncident = false) {
      if (!empty($idIncident) && is_numeric($idIncident)) {
        $data = Database::queryMulti('
              SELECT blck_incidents_status.id_incident_status, `title`,
                     `date_added`,
              DATE_FORMAT(`date_added`, "%e. %c. %Y %H:%i") AS `date_added`
              FROM `blck_incidents_status`
              LEFT JOIN
                  (SELECT `id_incident_status`, `date_added`
                   FROM `blck_incidents_status_log`
                   WHERE `id_incident` = ?)
              AS `log`
              ON blck_incidents_status.id_incident_status = log.id_incident_status
              WHERE `visible` = 1
              ORDER BY `id_incident_status`', $idIncident);
      }
      else {
        $data = Database::queryMulti('
              SELECT `id_incident_status`, `title`
              FROM `blck_incidents_status` 
              WHERE `visible` = 1');
      }

      return $data;
    }


    /**
     * Zvýrazní a barevně odliší v logu jednotlivé identifikátory předané v parametrech metody.
     *
     * @param string $ipAddress         IP adresa ke zvýraznění.
     * @param string[] $macAddresses    Pole MAC adres ke zvýraznění.
     * @param string[] $hostnames       Pole hostnames ke zvýraznění.
     * @param string[] $usernames       Pole uživatelských jmen ke zvýraznění.
     * @param string[] $log             Jednotlivé řádky logu, ve kterém se mají identifikátory zvýrazňovat.
     * @return string|null              Log převedený do podoby stringu se zvýrazněnými identifikátory nebo NULL.
     */
    private static function markIdentifiersInLogs($ipAddress, $macAddresses, $hostnames, $usernames, $log) {
      if (empty($log)) {
        return null;
      }

      if (is_array($log)) {
        $log = get_array_items_as_newline_string($log);
      }

      $markedLogs = str_replace(
        ["\n", "  ", $ipAddress, '[header]', '[/header]'],
        ['<br>', '&nbsp;&nbsp;', '<span style="color: purple">' . $ipAddress . '</span>', '<span class="text-body">', '</span>'],
        $log
      );

      // Zvýraznění všech předaných MAC adres.
      foreach ($macAddresses as $mac) {
        $markedLogs = str_replace($mac, '<span class="text-primary">' . $mac . '</span>', $markedLogs);
      }

      // Zvýraznění všech předaných hostnames.
      foreach ($hostnames as $hostname) {
        $markedLogs = str_replace($hostname, '<span class="text-success">' . $hostname . '</span>', $markedLogs);
      }

      // Zvýraznění všech předaných uživatelských jmen.
      foreach ($usernames as $username) {
        $markedLogs = str_replace($username, '<span class="text-danger">' . $username . '</span>', $markedLogs);
      }

      return $markedLogs;
    }


    /**
     * Vrátí CSS třídu v závislosti na rychlosti reakce řešení incidentu.
     *
     * @param int $reactionSpeed       Rychlost vyřešení incidentu (v hodinách).
     */
    private static function getReactionSpeedColor($reactionSpeed) {
      $cssClass = 'danger';

      if ($reactionSpeed < 12) {
        $cssClass = 'success';
      }
      else if ($reactionSpeed < 24) {
        $cssClass = 'warning text-dark';
      }

      return $cssClass;
    }


    /**
     * Zkontroluje uživatelský vstup (atributy třídy), který se bude zapisovat do databáze.
     *
     * @throws UserError               Výjimka obsahující textovou informaci o chybě pro uživatele.
     */
    public function validateData() {
      if ($this->currentStep == 1) {
        $this->isBlockedIPEmpty();
        $this->isBlockedIPValid();
        $this->isBlockedIPTooLong();
        $this->isBlockedIPFromValidRange();

        $this->isBlockedMACValid();
        $this->isBlockedMACTooLong();

        $this->isDatetimeEmpty();
        $this->isDatetimeValid();

        $this->isIncidentTypeEmpty();
        $this->existIncidentType();

        $this->isIdParentTicketValid();
      }
      else if ($this->currentStep == 2) {
        $this->isUserEmpty();
        $this->isUserValid();
        $this->isUserLangEmpty();
        $this->isUserLangValid();

        // Pokud byl v prvním kroku vyplněn lístek stěžovatele, pak
        // kontrolovat i vybrání jazyka pro komunikaci se stěžovatelem.
        if ($this->dbRecordData['id_parent_ticket'] != null) {
          $this->isParentTicketLangEmpty();
          $this->isParentTicketLangValid();
        }
      }
      else if ($this->currentStep == 3) {
        $this->isBlockEmailSubjectEmpty();
        $this->isBlockEmailBodyEmpty();
      }
    }


    /**
     * Ověří, zdali byla vyplněna IP adresa.
     *
     * @throws UserError               Výjimka obsahující textovou informaci o chybě pro uživatele.
     */
    private function isBlockedIPEmpty() {
      if (empty($this->blockedIp)) {
        throw new UserError('Není vyplněna IP adresa!', MSG_ERROR);
      }
    }


    /**
     * Ověří, zdali je zadaná IP adresa ve správném formátu.
     *
     * @throws UserError               Výjimka obsahující textovou informaci o chybě pro uživatele.
     */
    private function isBlockedIPValid() {
      if (filter_var($this->blockedIp, FILTER_VALIDATE_IP) === false) {
        throw new UserError('Zadaná IP adresa je v nesprávném formátu.', MSG_ERROR);
      }
    }


    /**
     * Ověří, zdali je zadaná IP adresa z rozsahu organizace, na které je nasazena aplikace Blockator.
     *
     * @throws UserError               Výjimka obsahující textovou informaci o chybě pro uživatele.
     */
    private function isBlockedIPFromValidRange() {
      if (filter_var($this->blockedIp, FILTER_VALIDATE_IP, FILTER_FLAG_IPV4) &&
        (ip2long($this->blockedIp) < ip2long(IP_ORG_START) || ip2long($this->blockedIp) > ip2long(IP_ORG_END))) {
        throw new UserError('Zadaná IP adresa není ve správě organizace.', MSG_ERROR);
      }
    }


    /**
     * Ověří, zdali zadaná IP adresa není příliš dlouhá.
     *
     * @throws UserError               Výjimka obsahující textovou informaci o chybě pro uživatele.
     */
    private function isBlockedIPTooLong() {
      if (mb_strlen($this->blockedIp) > $this->inputsMaxLengths['blocked-ip']) {
        throw new UserError('Zadaná IP adresa je příliš dlouhá.', MSG_ERROR);
      }
    }


    /**
     * Ověří, zdali je zadaná MAC adresa ve správném formátu.
     *
     * @throws UserError               Výjimka obsahující textovou informaci o chybě pro uživatele.
     */
    private function isBlockedMACValid() {
      if (!empty($this->blockedMac) && filter_var($this->blockedMac, FILTER_VALIDATE_MAC) === false) {
        throw new UserError('Zadaná MAC adresa je v nesprávném formátu.', MSG_ERROR);
      }
    }


    /**
     * Ověří, zdali zadaná MAC adresa není příliš dlouhá.
     *
     * @throws UserError               Výjimka obsahující textovou informaci o chybě pro uživatele.
     */
    private function isBlockedMACTooLong() {
      if (!empty($this->blockedMac) && mb_strlen($this->blockedMac) > $this->inputsMaxLengths['blocked-mac']) {
        throw new UserError('Zadaná MAC adresa je příliš dlouhá.', MSG_ERROR);
      }
    }


    /**
     * Ověří, zdali byl vyplněn datum a čas výskytu incidentu.
     *
     * @throws UserError               Výjimka obsahující textovou informaci o chybě pro uživatele.
     */
    private function isDatetimeEmpty() {
      if (empty($this->incidentDatetime)) {
        throw new UserError('Není vyplněn datum a čas výskytu incidentu!', MSG_ERROR);
      }
    }


    /**
     * Ověří, zdali vyplněné datum a čas výskytu incidentu není z budoucnosti.
     *
     * @throws UserError               Výjimka obsahující textovou informaci o chybě pro uživatele.
     */
    private function isDatetimeValid() {
      if (strtotime($this->incidentDatetime) > strtotime('now')) {
        throw new UserError('Vyplněný datum a čas výskytu incidentu je z data, které ještě nenastalo!', MSG_ERROR);
      }
    }


    /**
     * Ověří, zdali byl vybrán typ incidentu.
     *
     * @throws UserError               Výjimka obsahující textovou informaci o chybě pro uživatele.
     */
    private function isIncidentTypeEmpty() {
      if (empty($this->idType) || !is_numeric($this->idType)) {
        throw new UserError('Není vybrán typ incidentu.', MSG_ERROR);
      }
    }


    /**
     * Ověří, zdali vybraný typ incidentu existuje.
     *
     * @throws UserError               Výjimka obsahující textovou informaci o chybě pro uživatele.
     */
    private function existIncidentType() {
      $model = new IncidentTypeModel();

      if (empty($model->getIncidentType($this->idType))) {
        throw new UserError('Vybraný typ incidentu neexistuje.', MSG_ERROR);
      }
    }


    /**
     * Ověří, zdali je vyplněné číslo lístku se stížností ve správném formátu.
     *
     * @throws UserError               Výjimka obsahující textovou informaci o chybě pro uživatele.
     */
    private function isIdParentTicketValid() {
      if (!empty($this->idParentTicket) && (!is_numeric($this->idParentTicket) || $this->idParentTicket <= 0)) {
        throw new UserError('Číslo lístku se stížností je v nesprávném formátu!', MSG_ERROR);
      }
    }


    /**
     * Ověří, zdali byl vybrán uživatel, který má být zablokován.
     *
     * @throws UserError               Výjimka obsahující textovou informaci o chybě pro uživatele.
     */
    private function isUserEmpty() {
      if (empty($this->user)) {
        throw new UserError('Nebyl vybrán uživatel, který má být zablokován!', MSG_ERROR);
      }
    }


    /**
     * Ověří, zdali byl vybrán uživatel, který má být zablokován.
     *
     * @throws UserError               Výjimka obsahující textovou informaci o chybě pro uživatele.
     */
    private function isUserValid() {
      foreach ($this->user as $user) {
        if (!in_array($user, $this->dbRecordData['identifiedUsers'])) {
          throw new UserError('Vybraný uživatel není mezi těmi uživateli, kteři by mohli být zablokováni!', MSG_ERROR);
        }
      }
    }


    /**
     * Ověří, zdali byl vybrán jazyk, v němž bude probíhat komunikace s uživatelem.
     *
     * @throws UserError               Výjimka obsahující textovou informaci o chybě pro uživatele.
     */
    private function isUserLangEmpty() {
      if (empty($this->userLang)) {
        throw new UserError('Nebyl vybrán jazyk, v němž bude probíhat komunikace s uživatelem!', MSG_ERROR);
      }
    }


    /**
     * Ověří, zdali vybraný jazyk pro komunikaci s uživatelem skutečně existuje.
     *
     * @throws UserError               Výjimka obsahující textovou informaci o chybě pro uživatele.
     */
    private function isUserLangValid() {
      $languages = EmailTemplatesModel::getEmailTemplatesLanguages($this->dbRecordData['id_type']);

      foreach ($this->userLang as $language) {
        if (!in_array($language, $languages)) {
          throw new UserError('Pro jazyk vybraný pro komunikaci s uživatelem neexistuje žádná šablona!', MSG_ERROR);
        }
      }
    }


    /**
     * Ověří, zdali byl vybrán jazyk, v němž bude probíhat komunikace se stěžovatelem.
     *
     * @throws UserError               Výjimka obsahující textovou informaci o chybě pro uživatele.
     */
    private function isParentTicketLangEmpty() {
      if (empty($this->parentTicketLang)) {
        throw new UserError('Nebyl vybrán jazyk, v němž bude probíhat komunikace se stěžovatelem!', MSG_ERROR);
      }
    }


    /**
     * Ověří, zdali vybraný jazyk pro komunikaci se stěžovatelem skutečně existuje.
     *
     * @throws UserError               Výjimka obsahující textovou informaci o chybě pro uživatele.
     */
    private function isParentTicketLangValid() {
      $languages = EmailTemplatesModel::getEmailTemplatesLanguages($this->dbRecordData['id_type']);

      foreach ($this->parentTicketLang as $language) {
        if (!in_array($language, $languages)) {
          throw new UserError('Pro jazyk vybraný pro komunikaci se stěžovatelem neexistuje žádná šablona!', MSG_ERROR);
        }
      }
    }


    /**
     * Ověří, zdali byl vyplněn předmět zprávy o blokaci.
     *
     * @throws UserError               Výjimka obsahující textovou informaci o chybě pro uživatele.
     */
    private function isBlockEmailSubjectEmpty() {
      if (empty($this->blockingSubject)) {
        throw new UserError('Není vyplněn obsah zprávy o blokaci!', MSG_ERROR);
      }
    }


    /**
     * Ověří, zdali byl vyplněn obsah zprávy o blokaci.
     *
     * @throws UserError               Výjimka obsahující textovou informaci o chybě pro uživatele.
     */
    private function isBlockEmailBodyEmpty() {
      if (empty($this->blockingMessage)) {
        throw new UserError('Není vyplněn obsah zprávy o blokaci!', MSG_ERROR);
      }
    }
  }
