<?php
  /**
   * Třída zajišťující shromáždění dat pro grafy, statistiku a související operace.
   *
   * @author Martin Šebela
   */
  class StatsModel {
    /**
     * Připraví a vrátí data pro graf znázorňující aktuální rozložení
     * stavů u všech evidovaných bezpečnostních incidentů.
     *
     * Výstupem je pole o následujících indexech:
     * - legend = data pro legendu,
     * - colors = barvy legendy,
     * - data = konkrétní počty.
     *
     * @return array                   Pole s daty pro graf.
     */
    public static function getIncidentStatusesChart() {
      $chart = [];

      $data = Database::queryMulti('
        SELECT * FROM
            (SELECT DISTINCT blck_incidents.id_incident, blck_incidents_status_log.id_incident_status,
            `title`, `color`
            FROM `blck_incidents`
            JOIN `blck_incidents_status_log`
            ON blck_incidents.id_incident = blck_incidents_status_log.id_incident
            JOIN `blck_incidents_status`
            ON blck_incidents_status_log.id_incident_status = blck_incidents_status.id_incident_status 
            WHERE blck_incidents.visible = 1
            ORDER BY `id_incident_status` DESC) AS `last_statuses`
        GROUP BY `id_incident`
      ');

      foreach ($data as $record) {
        $type = $record['id_incident_status'];

        // Pokud je to první výskyt dané hodnoty.
        if (!isset($chart['data'][$type])) {
          $chart['data'][$type] = 1;
        }
        else {
          $chart['data'][$type]++;
        }

        // Přidání názvu a barvy.
        $chart['legend'][$type] = $record['title'];
        $chart['colors'][$type] = $record['color'];
      }

      // Transformace dat pro potřeby knihovny Chart.js.
      if (!empty($data)) {
        $chart = self::transformData($chart);
      }

      return $chart;
    }


    /**
     * Připraví a vrátí data pro graf znázorňující všechny
     * evidované bezpečnostní incidenty dle jejich typu.
     *
     * Výstupem je pole o následujících indexech:
     * - legend = data pro legendu,
     * - colors = barvy legendy,
     * - data = konkrétní počty.
     *
     * @return array                   Pole s daty pro graf.
     */
    public static function getIncidentTypesChart() {
      $chart = [];

      $data = Database::queryMulti('
        SELECT `id_type`,
        `name`, `color`
        FROM `blck_incidents`
        JOIN `blck_incidents_types`
        ON blck_incidents.id_type = blck_incidents_types.id_incident_type 
        WHERE blck_incidents.visible = 1
      ');

      foreach ($data as $record) {
        $type = $record['id_type'];

        // Pokud je to první výskyt dané hodnoty.
        if (!isset($chart['data'][$type])) {
          $chart['data'][$type] = 1;
        }
        else {
          $chart['data'][$type]++;
        }

        $chart['legend'][$type] = $record['name'];
        $chart['colors'][$type] = $record['color'];
      }

      // Transformace dat pro potřeby knihovny Chart.js.
      if (!empty($data)) {
        $chart = self::transformData($chart);
      }

      return $chart;
    }


    /**
     * Připraví a vrátí data pro graf znázorňující všechny řešené bezpečnostní
     * incidenty dle jejich typu v jednotlivých měsících a v konkrétním roku.
     *
     * @param int $year                Rok, který má být uvažován.
     * @return array                   Pole s daty pro graf.
     */
    public static function getIncidentsTypesByYear($year) {
      $datasets = [];

      $yearData = Database::queryMulti('
        SELECT YEAR(`incident_datetime`) AS `year`, MONTH(`incident_datetime`) AS `month`, `id_type`,
        `name`, `color`
        FROM `blck_incidents`
        JOIN `blck_incidents_types`
        ON blck_incidents.id_type = blck_incidents_types.id_incident_type 
        WHERE blck_incidents.visible = 1
        ORDER BY YEAR(`incident_datetime`), MONTH(`incident_datetime`)');

      foreach ($yearData as &$month) {
        $foundDataset = false;

        // Zjištění, zdali je daný typ incidentu již uložen v některém z datasetů.
        foreach ($datasets as $record) {
          if (isset($record['title']) && strcmp($record['title'], $month['name']) == 0) {
            $foundDataset = true;
            break;
          }
        }

        // Vytvoření nového datasetu s novým typem incidentu.
        if (!$foundDataset) {
          $months = self::getEmptyMonthsArray();
          $months[$month['month']]++;
          $datasets[] = ['title' => $month['name'], 'color' => $month['color'], 'data' => $months];
        }
        else {
          // Dataset s tímto typem incidentu již existuje - inkrementovat tedy pouze výskyt u správného měsíce.
          foreach ($datasets as &$recordd) {
            if (strcmp($recordd['title'], $month['name']) == 0) {
              $recordd['data'][$month['month']]++;
            }
          }
        }
      }

      return $datasets;
    }


    /**
     * Změní pole na řetězec tak, že jednotlivé hodnoty původního pole oddělí zvoleným oddělovačem
     * a případně ještě před danou hodnotu přidá zvolený prefix (nepovinný parametr).
     *
     * @param array $array             Pole, které se má transformovat na řetězec.
     * @param string $delimiter        Oddělovač hodnot pole.
     * @param string $valuePrefix      Prefix všech hodnot pole.
     * @return string                  Řetězec vzniklý z původního pole.
     */
    private static function getArrayAsDelimiterString($array, $delimiter, $valuePrefix = '') {
      $string = '';

      if (!empty($array)) {
        foreach ($array as $value) {
          $string .= $delimiter . $valuePrefix . $value . $delimiter . ', ';
        }
      }

      // Odstranění oddělovače u poslední položky, která byla do řetězce přidána.
      return rtrim($string, ', ');
    }


    /**
     * Vrátí prázdné pole o dvanácti prvcích reprezentujících jednotlivé měsíce v roce.
     *
     * @return array                   Prázdné pole o 12 prvcích.
     */
    private static function getEmptyMonthsArray() {
      $months = [];

      for ($i = 1; $i <= 12; $i++) {
        $months[$i] = 0;
      }

      return $months;
    }


    /**
     * Vrátí české názvy měsíců v roce jako řetězec pro legendu.
     *
     * @return string                  České názvy měsíců.
     */
    public static function getMonthsNames() {
      $monthsNames = [
        'leden', 'únor', 'březen', 'duben', 'květen', 'červen', 'červenec', 'srpen', 'září', 'říjen', 'listopad', 'prosinec'
      ];

      return self::getArrayAsDelimiterString($monthsNames, "'");
    }

    /**
     * Transformuje pole s daty pro koláčové grafy do podoby požadované knihovnou Chart.js.
     *
     * @param array $chart             Původní data pro graf formou pole.
     * @return array                   Transformovaná data v poli.
     */
    private static function transformData($chart) {
      if (isset($chart['data'])) {
        $chart['data'] = implode($chart['data'], ',');
      }

      if (isset($chart['legend'])) {
        $chart['legend'] = self::getArrayAsDelimiterString($chart['legend'], '"');
      }

      if (isset($chart['colors'])) {
        $chart['colors'] = self::getArrayAsDelimiterString($chart['colors'], '"');
      }

      return $chart;
    }
  }
