<?php
  /**
   * Abstraktní třída Blockátoru, jejíž metody dědí Blockátory (moduly) spravující jednotlivé části sítě.
   * Třída poskytuje metody, které jsou užitečné napříč různými Blockátory (moduly).
   *
   * @author Martin Šebela
   */
  abstract class Blocker {
    /**
     * Vrátí název (označení) sítě pro textové výpisy.
     *
     * @return string                  Název sítě.
     */
    public function getNetworkName() {
      return $this->networkName;
    }


    /**
     * Vloží poznámku ke konkrétnímu užiteli (např. o blokaci, odblokování) do agendního systému.
     *
     * @param string $username         Uživatelské jméno, jehož se poznámka týká.
     * @param string $note             Poznámka do agendního systému k danému uživateli.
     * @return bool                    TRUE pokud byla poznámka do agendního systému úspěšně přidána.
     * @throws UserError               Výjimka obsahující textovou informaci o chybě pro uživatele.
     */
    public function addNoteToUsersDatabase($username, $note) {
      if (strpos($username, '@') !== false) {
        $username = get_email_part($username, 'username');
      }

      $arguments = [
        'login' => $username,
        'pozn' => $note
      ];

      $url = 'https://whois3.zcu.cz/web1/robot?' . http_build_query($arguments);
      $response = process_sso($url);

      if ($response == '1') {
        return true;
      }
      else {
        throw new UserError('Nepodařilo se vytvořit spojení se vzdálenou službu WhoIS.', MSG_ERROR);
      }
    }


    /**
     * Vrátí textovou zprávu o blokaci pro agendní systémy.
     *
     * @param BlockedIdentity $identity Identita, jíž se blokace týká.
     * @return string                   Textová zpráva o blokaci.
     */
    protected function getBlockedMessage(BlockedIdentity $identity) {
      return 'Blockator: Blokace ' . $this->getNetworkName() . ' z důvodu ' . $identity->getReason() . ' (RT ' . $identity->getIdRequestTracker() . ')';
    }


    /**
     * Vrátí textovou zprávu o blokaci pro RT systém (systém pro správu požadavků).
     *
     * @param BlockedIdentity $identity Identita, jíž se blokace týká.
     * @return string                   Textová zpráva o blokaci.
     */
    public function getRTBlockedMessage(BlockedIdentity $identity) {
      return 'Blockator: Blokace ' . $this->getNetworkName() . ' (účet ' . $identity->getIdentity() . ') z důvodu ' . $identity->getReason();
    }


    /**
     * Vrátí textovou zprávu o odblokování pro agendní systémy.
     *
     * @param BlockedIdentity $identity Identita, jíž se odblokování týká.
     * @return string                   Textová zpráva o odblokování.
     */
    protected function getUnblockedMessage(BlockedIdentity $identity) {
      return 'Blockator: Odblokování ' . $this->getNetworkName() . ' (RT ' . $identity->getIdRequestTracker() . ')';
    }
  }