<?php
  /**
   * Blockátor pro Eduroam síť. Obsahuje funkce pro vyhledání uživatele z logů,
   * funkce pro blokaci a odblokování konkrétního uživatele.
   *
   * @author Martin Šebela
   */
  class EduroamBlocker extends Blocker implements IBlocker {
    /**
     * @var string      Název (označení) sítě pro textové výpisy.
     */
    protected $networkName = 'eduroam';

    /**
     * @var string[]    Vzor na základě něhož bude rozhodnuto, že zařízení spadá do rozsahu této sítě.
     */
    private $networkPatternHostname = ['eduroam', 'zcu-mobile'];

    /**
     * @var string      URL volané služby, kde dochází k blokaci nebo odblokování.
     */
    private $serviceBaseUrl = 'https://smp.zcu.cz/smp/MySphere';

    /**
     * @var string      Cesta k bash skriptu, který slouží k (od)blokaci uživatelů z jiných RADIUS serverů (organizací).
     */
    private $scriptRadiusFilepath = CORE_DOCUMENT_ROOT . '/' . CORE_DIR_BLOCKERS . '/remoteServers.sh';

    /**
     * @var string      Cesta v souborovém systému, kde jsou uloženy DHCP logy.
     */
    private $logBaseFilepath = '/mnt/data/dhcpd-zcu/';


    /**
     * Vrátí, zdali hostname spadá do rozsahu sítě.
     *
     * @param string $hostname         Hostname zařízení
     * @return bool                    TRUE pokud spadá do oblasti sítě, jinak FALSE.
     */
    public function isFromThisNetwork($hostname) {
      foreach ($this->networkPatternHostname as $networkHostname) {
        if (strpos($hostname, mb_strtolower($networkHostname)) !== false) {
          return true;
        }
      }

      return false;
    }


    /**
     * Zařídí vyhledávání záznamů v logu o dané IP adrese v určitém datu a čase a nalezená data vrátí.
     *
     * @param string $ipAddress        IP adresa
     * @param string $macAddress       MAC adresa
     * @param string $datetime         Datum a čas
     * @return array                   Nalezená data v logu
     * @throws UserError               Výjimka obsahující textovou informaci o chybě pro uživatele.
     */
    public function find($ipAddress, $macAddress, $datetime) {
      $foundData = $this->searchInDhcpLog($ipAddress, $macAddress, $datetime);

      return [
        'logs' => $foundData['logs'],
        'macAddresses' => $foundData['macAddresses'],
        'hostnames' => $foundData['hostnames'],
        'users' => $foundData['users']
      ];
    }


    /**
     * Zablokuje konkrétního uživatele v síti Eduroam.
     *
     * @param BlockedIdentity $identity Identita, která se má zablokovat ve službě SMP.
     * @return bool                     TRUE při úspěšné blokaci, jinak FALSE.
     * @throws UserError                Výjimka obsahující textovou informaci o chybě pro uživatele.
     */
    public function block(BlockedIdentity $identity) {
      $result = false;

      if (is_user_from_organization($identity->getIdentity())) {
        $note = $this->getBlockedMessage($identity);

        $arguments = [
          'login' => $identity->getUsername(),
          'duvod' => $note,
          //'exp' => '', (případné datum expirace)
          'inc_num' => $identity->getIdIncident(),
          'rt_num' => $identity->getIdRequestTracker(),
          'action' => 'block',
          'xsubmit' => 'Zablokovat!'
        ];

        // Uživatel úspěšně zablokován.
        if ($this->callService($arguments)) {
          $this->addNoteToUsersDatabase($identity->getUsername(), $note);
          $result = true;
        }
      }
      else {
        // Pokud je uživatel z jiné organizace...
        $result = $this->blockOutsideOrg($identity);
      }

      return $result;
    }


    /**
     * Zablokuje uživatele z jiné organizace v síti Eduroam.
     *
     * @param BlockedIdentity $identity Identita, která se má zablokovat.
     * @return bool                     TRUE při úspěšné blokaci.
     * @throws UserError                Výjimka obsahující textovou informaci o chybě pro uživatele.
     */
    private function blockOutsideOrg(BlockedIdentity $identity) {
      $result = false;

      if (is_valid_username($identity->getIdentity())) {
        $username = escapeshellcmd($identity->getIdentity());
        shell_exec($this->scriptRadiusFilepath . ' eduroam block "' . $username . '"');

        $result = true;
      }

      return $result;
    }


    /**
     * Odblokuje konkrétního uživatele v síti Eduroam.
     *
     * @param BlockedIdentity $identity Identita, která se má odblokovat ve službě SMP.
     * @return bool                     TRUE při úspěšném odblokování, jinak FALSE.
     * @throws UserError                Výjimka obsahující textovou informaci o chybě pro uživatele.
     */
    public function unblock(BlockedIdentity $identity) {
      $result = false;

      if (is_user_from_organization($identity->getIdentity())) {
        $note = $this->getUnblockedMessage($identity);

        $arguments = [
          'login' => $identity->getUsername(),
          'duvod' => $note,
          'action' => 'unblock',
          'xsubmit' => 'Odblokovat!'
        ];

        // Uživatel úspěšně odblokován.
        if ($this->callService($arguments)) {
          $this->addNoteToUsersDatabase($identity->getUsername(), $note);
          $result = true;
        }
      }
      else {
        // Pokud je uživatel z jiné organizace...
        $result = $this->unblockOutsideOrg($identity);
      }

      return $result;
    }


    /**
     * Odblokuje uživatele z jiné organizace v síti Eduroam.
     *
     * @param BlockedIdentity $identity Identita, která se má zablokovat.
     * @return bool                     TRUE při úspěšném odblokování.
     * @throws UserError                Výjimka obsahující textovou informaci o chybě pro uživatele.
     */
    private function unblockOutsideOrg(BlockedIdentity $identity) {
      $result = false;

      if (is_valid_username($identity->getIdentity())) {
        $username = escapeshellcmd($identity->getIdentity());
        shell_exec($this->scriptRadiusFilepath . ' eduroam unblock "' . $username . '"');

        $result = true;
      }

      return $result;
    }


    /**
     * Zavolá externí službu SMP pro blokaci/odblokování uživatelů a předá jí požadované HTTP argumenty.
     *
     * @param array $arguments         Argumenty pro HTTP dotaz.
     * @return bool                    Vrátí TRUE při úspěšném volání služby a při úspěšném provedení akce.
     * @throws UserError               Výjimka obsahující textovou informaci o chybě pro uživatele.
     */
    private function callService($arguments) {
      $url = $this->serviceBaseUrl . '?' . http_build_query($arguments);
      $response = process_sso($url);
      $data = json_decode($response, true);

      // Volání služby proběhlo úspěšně.
      if ($data['ret_code'] == 'OK') {
        return true;
      }
      // Chyba, kterou vrátíla externí služba SMP.
      else if ($data['ret_code'] == 'ERROR') {
        throw new UserError('Došlo k chybě ve vzdálené službě SMP: ' . Controller::escapeOutput($data['ret_msg']), MSG_ERROR);
      }
      // Spojení se službou se nepodařilo (např. volání neprošlo přes SSO).
      else {
        throw new UserError('Nepodařilo se vytvořit spojení se vzdálenou službou SMP.', MSG_ERROR);
      }
    }


    /**
     * Vyhledá záznamy v DHCP logu o dané IP adrese a v inkriminovaném datu a čase.
     *
     * @param string $ipAddress        IP adresa
     * @param string $macAddress       MAC adresa zařízení (nepovinné)
     * @param string $datetime         Datum a čas
     * @return array                   Záznamy nalezené v logu
     * @throws UserError               Výjimka obsahující textovou informaci o chybě pro uživatele.
     */
    private function searchInDhcpLog($ipAddress, $macAddress, $datetime) {
      $logRecords = [];
      $macAddresses = [];
      $hostnames = [];
      $foundUsers = [];

      $date = get_date_parts($datetime);

      // Cesta k DHCP logu pro dané datum.
      $file = $this->logBaseFilepath . $date['year'] . '/' . $date['month'] . '/dhcpd.log.' . $date['year'] . $date['month'] . $date['day'];

      $contents = file_get_contents($file);

      // Regulární výraz pro vyhledávání v logu.
      if ($macAddress != null) {
        // Vyhledávání podle IP i MAC adresy.
        $regexp = '(.*) DHCPACK on ' . $ipAddress . ' to ' . $macAddress . ' \((.*)\) via (.*)';
      }
      else {
        // Vyhledávání pouze podle IP adresy.
        $regexp = '(.*) DHCPACK on ' . $ipAddress . ' to (.*) \((.*)\) via (.*)';
      }

      // Nalezení záznamů DHCPACK s hledanou IP adresou v inkriminovaném datu.
      if (preg_match_all('/' . $regexp . '/', $contents, $matches)) {
        $logRecords[] = '[header]--------------------------- DHCP log ---------------------------[/header]';
        $logRecords = array_merge($logRecords, $matches[0]);

        if ($macAddress != null) {
          $macAddresses = [$macAddress];
          $hostnames = $matches[2];
        }
        else {
          $macAddresses = $matches[2];
          $hostnames = $matches[3];
        }
      }

      // Odstranění duplicit.
      $macAddresses = remove_duplicates_in_array($macAddresses);
      $hostnames = remove_duplicates_in_array($hostnames);

      // Vyhledání nalezené MAC adresy v RADIUS logu a vyhledání konkrétního uživatele.
      foreach ($macAddresses as $mac) {
        $result = $this->searchInRadiusLog($mac, $datetime, 0);

        $logRecords[] = "\n\n" . $result['logs'];

        foreach ($result['users'] as $user) {
          $foundUsers[] = $user;
        }
      }

      return [
        'logs' => $logRecords,
        'macAddresses' => $macAddresses,
        'hostnames' => $hostnames,
        'users' => $foundUsers
      ];
    }


    /**
     * Vyhledá v RADIUS logu záznamy o konkrétní IP/MAC adrese ve zvoleném čase.
     *
     * @param string $userIdentifier   IP/MAC adresa, která se má v logu vyhledávat.
     * @param string $datetime         Datum a čas, které se má během vyhledávání brát v úvahu.
     * @param int $secondsShift        Úprava počátečního času vyhledávání.
     * @return array                   Informace (tj. logy a konkrétní uživatelé) nalezení v logu vrácené formou pole.
     * @throws UserError               Výjimka obsahující textovou informaci o chybě pro uživatele.
     */
    private function searchInRadiusLog($userIdentifier, $datetime, $secondsShift) {
      $foundUsers = [];

      $radiusConnection = $this->connectToRadiusDatabase();

      $timestamp = strtotime($datetime);
      $startTime = strftime('%Y-%m-%d %H:%M:00', $timestamp + $secondsShift);
      $endTime = strftime('%Y-%m-%d %H:%M:00', $timestamp);

      $cols = ['username', 'framedipaddress', 'acctsessionid', 'acctstarttime', 'acctstoptime', 'acctsessiontime', 'acctterminatecause', 'radacctid'];

      // Dodatečné ošetření před vykonáním SQL dotazu.
      $userIdentifier = mysqli_real_escape_string($radiusConnection, $userIdentifier);
      $startTime = mysqli_real_escape_string($radiusConnection, $startTime);
      $endTime = mysqli_real_escape_string($radiusConnection, $endTime);

      if (filter_var($userIdentifier, FILTER_VALIDATE_IP) !== false) {
        // Vyhledávání v databázi podle použité IP adresy.
        $whereFilter = 'framedipaddress = "' . $userIdentifier . '" AND acctstarttime <= "' . $startTime . '" AND acctstoptime >= "' . $endTime . '"';
      }
      else if (filter_var($userIdentifier, FILTER_VALIDATE_MAC) !== false) {
        // Vyhledávání v databázi podle použité MAC adresy.
        $whereFilter = 'acctstarttime >= DATE_SUB("' . $startTime . '", INTERVAL 7 DAY) AND acctsessionid LIKE "%' . $userIdentifier . '%" GROUP BY username, DATE(acctstarttime)';
      }

      // Vyhledávání všech záznamů o připojení dané IP/MAC adresy v inkriminovaném čase v RADIUS logu.
      $query = 'SELECT ' . implode(',', $cols) . ' FROM radacct WHERE ' . $whereFilter;

      $colsHeaderFormat = "%15s %20s %35s %20s %20s %16s %19s\n";
      $logRecords = '[header]--------------------------- RADIUS log ---------------------------[/header]' . "\n";
      $logRecords .= '[header]' . vsprintf($colsHeaderFormat, $cols) . '[/header]';

      $result = mysqli_query($radiusConnection, $query);

      while ($record = mysqli_fetch_assoc($result)) {
        $logRecords .= vsprintf($colsHeaderFormat, $record);
      }

      // Vyhledání konkrétního uživatele v RADIUS logu (bez duplicit).
      $query = 'SELECT DISTINCT(username) FROM radacct WHERE ' . $whereFilter;

      $result = mysqli_query($radiusConnection, $query);

      while ($record = mysqli_fetch_assoc($result)) {
        $foundUsers[] = $record['username'];
      }

      return [
        'logs' => $logRecords,
        'users' => $foundUsers
      ];
    }


    /**
     * Vytvoří spojení s MySQL RADIUS databází uživatelů a vytvořené spojení vrátí.
     *
     * @return mysqli                  Databázové spojení s MySQL databází RADIUS serveru
     * @throws UserError               Výjimka obsahující textovou informaci o chybě pro uživatele.
     */
    private function connectToRadiusDatabase() {
      $servername = '';
      $username = '';
      $password = '';
      $databaseName = '';
      $port = '';

      $connection = mysqli_connect($servername, $username, $password, $databaseName, $port);

      if (!$connection) {
        throw new UserError('Připojení k RADIUS databázi se nepodařilo: ' . mysqli_connect_error(), MSG_ERROR);
      }

      return $connection;
    }
  }